package edu.berkeley.eas.enrollment.dao.sreservation;

import java.util.List;
import java.util.Map;
import javax.sql.DataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import edu.berkeley.eas.enrollment.exception.EnrollmentServiceException;
import edu.berkeley.eas.enrollment.model.common.CourseInfo;
import edu.berkeley.eas.enrollment.model.common.UgGradCodeEnum;
import edu.berkeley.eas.enrollment.model.sreservation.CodesInfo;
import edu.berkeley.eas.enrollment.model.sreservation.SectionReservations;
import edu.berkeley.eas.enrollment.model.sreservation.SectionReservations.ReservationInfo;
import edu.berkeley.eas.enrollment.model.sreservation.SummaryInfo;
import edu.berkeley.eas.enrollment.util.CodeNamePairBuilder;
import edu.berkeley.eas.enrollment.util.CourseKeyBuilder;
import edu.berkeley.eas.enrollment.util.DepartmentInfoBuilder;
import edu.berkeley.eas.enrollment.util.ErrorInfoBuilder;
import edu.berkeley.eas.enrollment.util.SectionKeyBuilder;
import edu.berkeley.eas.enrollment.util.StrUtils;
import edu.berkeley.eas.enrollment.util.TermKeyBuilder;

import static edu.berkeley.eas.enrollment.constants.Constants.INVALID_CCN_MSG;
import static edu.berkeley.eas.enrollment.constants.Constants.INVALID_TERM_CODE_MSG;
import static edu.berkeley.eas.enrollment.constants.Constants.PROC_FAILED_MSG;
import static edu.berkeley.eas.enrollment.constants.Constants.PROC_NAME_MSG;
import static edu.berkeley.eas.enrollment.constants.Constants.RESULT_SET_PARAM;
import static edu.berkeley.eas.enrollment.constants.Constants.SECTION_RESERVATIONS_PROC_PROPERTY;

/**
 * SectionReservationsDao provides the Spring implementation for the Section
 * Reservations Rest Service
 *
 * @author khensel@berkeley.edu
 * @version %I%, %G%
 */
public class SectionReservationsDao {

    private static final Logger LOGGER = LoggerFactory.getLogger(SectionReservationsDao.class);

    private String procName = System.getProperty(SECTION_RESERVATIONS_PROC_PROPERTY);

    public void setProcName(String procName) {
        this.procName = procName;
    }

    private DataSource dataSource;

    public void setDataSource(DataSource dataSource) {
        this.dataSource = dataSource;
    }

    /**
     * Gets SectionReservations for ccn, term year, term code.
     *
     * @param ccn      Five digit Course Control Number
     * @param termYear Four digit Term Year
     * @param termCode B: Spring, D: Fall, C: Summer
     * @return SectionReservations XML Object of section reservations from schema
     * @throws EnrollmentServiceException System error occurred.
     */
    @SuppressWarnings("unchecked")
    public SectionReservations getSectionReservations(String ccn, String termYear, String termCode)
            throws EnrollmentServiceException {

        LOGGER.debug("+++ in SectionReservationsDao.getSectionReservations({}, {}, {})",
                ccn, termYear, termCode);
        LOGGER.debug(PROC_NAME_MSG, procName);

        SectionReservations rval = new SectionReservations();
        rval.setCourseInfo(new CourseInfo());

        try {
            rval.setTermKey(TermKeyBuilder.create(termYear, termCode));
            termCode = termCode.toUpperCase();
        } catch (IllegalArgumentException e) {
            rval.setErrorInfo(ErrorInfoBuilder.create(String.format(INVALID_TERM_CODE_MSG, termCode)));
            return rval;
        }

        int iccn;
        try {
            iccn = StrUtils.isEmpty(ccn) ? 0 : Integer.parseInt(ccn);
            rval.setCourseInfo(new CourseInfo());
            rval.getCourseInfo().setCourseControlNumber(StrUtils.padCCN(Integer.toString(iccn)));
        } catch (NumberFormatException e) {
            rval.setErrorInfo(ErrorInfoBuilder.create(String.format(INVALID_CCN_MSG, ccn)));
            return rval;
        }

        Map<String, Object> procParamMap;

        try {
            SectionReservationsProc storedProc = new SectionReservationsProc(dataSource, procName);
            procParamMap = storedProc.execute(iccn, termYear, termCode);
        } catch (Exception e) {
            LOGGER.error("+++ Exception occurred in Stored Proc: {} ccn: {} termYear: {} termCode: {}",
                    procName, ccn, termYear, termCode);
            throw new EnrollmentServiceException(e);
        }

        int returnCode = (Integer) procParamMap.get(SectionReservationsProc.SQL_RETURN_PARAM);
        String errorMessage = (String) procParamMap.get(SectionReservationsProc.ERROR_PARAM);

        if (returnCode > 4) {
            throw new EnrollmentServiceException(String.format(PROC_FAILED_MSG, procName, returnCode));
        }

        if (returnCode > 0) {
            rval.setErrorInfo(ErrorInfoBuilder.create(StrUtils.trim(errorMessage)));
            return rval;
        }

        List<SectionReservationsRowMapper> mapperList =
                (List<SectionReservationsRowMapper>) procParamMap.get(RESULT_SET_PARAM);

        if (mapperList == null || mapperList.isEmpty()) {
            rval.setErrorInfo(ErrorInfoBuilder.create("No sections found."));
            return rval;
        }

        // Element 0 contains restriction info and will always exist.
        SectionReservationsRowMapper mapper = mapperList.get(0);
        rval.getTermKey().setSubTermCode(mapper.getSubTermCode());
        rval.getCourseInfo().setDepartmentInfo(
                DepartmentInfoBuilder.create(mapper.getDepartmentCode(), mapper.getDepartmentName()));
        rval.getCourseInfo().setCourseKey(CourseKeyBuilder.create(
                mapper.getPrefix(), mapper.getRootNumber(), mapper.getSuffixOne(), mapper.getSuffixTwo()));
        rval.getCourseInfo().setSectionKey(SectionKeyBuilder.create(
                mapper.getPrimarySectionFlag(), mapper.getSectionNumber(), mapper.getInstructionalFormat()));

        SummaryInfo sectionSummary = getSummary(mapper);
        /*
           * instructorApprovalRequired is only carried on sectionSummaryInfo not
           * reservationSummary
           */
        sectionSummary.setInstructorApprovalReqired(mapper.isInstructorApprovalRequired());
        rval.setSectionSummary(sectionSummary);

        if (mapper.isCodesExist()) {
            rval.setRestrictionCodes(getCodes(mapper));
        }

        if (mapperList.size() > 1) {
            for (int i = 1; i < mapperList.size(); i++) {
                ReservationInfo reservationInfo = new ReservationInfo();
                mapper = mapperList.get(i);
                reservationInfo.setReservationSummary(getSummary(mapper));
                if (mapper.isCodesExist()) {
                    CodesInfo codes = getCodes(mapper);
                    reservationInfo.setCodes(codes);
                    rval.getReservationInfo().add(reservationInfo);
                }
            }
        }
        return rval;
    }

    private CodesInfo getCodes(SectionReservationsRowMapper mapper) throws EnrollmentServiceException {
        CodesInfo codes = new CodesInfo();
        codes.setSeatRestrictionCode(mapper.getSeatRestrictionCode());
        if (!StrUtils.isEmpty(mapper.getUgGradCode())) {
            try {
                UgGradCodeEnum ug = UgGradCodeEnum.fromValue(mapper.getUgGradCode());
                codes.setUgGradCode(ug);
            } catch (IllegalArgumentException e) {
                LOGGER.warn("+++ Invalid ugGradCode: '{}' for CCN: {} -- defaulting to '{}'",
                        mapper.getUgGradCode(), mapper.getCcn(), UgGradCodeEnum.NOT_SET.value());
                codes.setUgGradCode(UgGradCodeEnum.NOT_SET);
            }
        }

        codes.setCollegeOne(CodeNamePairBuilder.create(
                mapper.getCollegeCodeOne(), mapper.getCollegeNameOne()));
        codes.setCollegeTwo(CodeNamePairBuilder.create(
                mapper.getCollegeCodeTwo(), mapper.getCollegeNameTwo()));
        codes.setMajorOne(CodeNamePairBuilder.create(
                mapper.getMajorCodeOne(), mapper.getMajorNameOne()));
        codes.setMajorTwo(CodeNamePairBuilder.create(
                mapper.getMajorCodeTwo(), mapper.getMajorNameTwo()));
        codes.setMajorThree(CodeNamePairBuilder.create(
                mapper.getMajorCodeThree(), mapper.getMajorNameThree()));
        codes.setTelebearsLevelOne(CodeNamePairBuilder.create(
                mapper.getEducationLevelOne(), mapper.getEducationLevelOneDesc()));
        codes.setTelebearsLevelTwo(CodeNamePairBuilder.create(
                mapper.getEducationLevelTwo(), mapper.getEducationLevelTwoDesc()));
        codes.setTelebearsLevelThree(CodeNamePairBuilder.create(
                mapper.getEducationLevelThree(), mapper.getEducationLevelThreeDesc()));
        codes.setEligRegStatusOne(CodeNamePairBuilder.create(
                mapper.getRegStatusOne(), mapper.getRegStatusOneDesc()));
        codes.setEligRegStatusTwo(CodeNamePairBuilder.create(
                mapper.getRegStatusTwo(), mapper.getRegStatusTwoDesc()));
        codes.setUcRequirementOne(CodeNamePairBuilder.create(
                mapper.getUcRequirementOne(), mapper.getUcRequirementOneDesc()));
        codes.setUcRequirementTwo(CodeNamePairBuilder.create(
                mapper.getUcRequirementTwo(), mapper.getUcRequirementTwoDesc()));

        if (!StrUtils.isEmpty(mapper.getEducationLevelType()))
            codes.setEducationLevelType(mapper.getEducationLevelType());
        if (!StrUtils.isEmpty(mapper.getEnglishCompRequiredCode()))
            codes.setEnglishCompRequiredCode(mapper.getEnglishCompRequiredCode());
        if (!StrUtils.isEmpty(mapper.getAdmitSpecialProgramCodeOne()))
            codes.setAdmitSpecialProgramCodeOne(mapper.getAdmitSpecialProgramCodeOne());
        if (!StrUtils.isEmpty(mapper.getAdmitSpecialProgramCodeTwo()))
            codes.setAdmitSpecialProgramCodeTwo(mapper.getAdmitSpecialProgramCodeTwo());

        if (!StrUtils.isEmpty(mapper.getPriorityPattern()))
            codes.setPriorityPattern(mapper.getPriorityPattern());

        return codes;
    }

    private SummaryInfo getSummary(SectionReservationsRowMapper mapper) {
        SummaryInfo summary = new SummaryInfo();
        summary.setFull(mapper.isFull());
        summary.setEnrollCount(mapper.getEnrollCount());
        summary.setEnrollLimit(mapper.getEnrollLimit());
        summary.setWaitListCode(mapper.getWaitListCode());
        summary.setWaitListCount(mapper.getWaitListCount());
        summary.setWaitListLimit(mapper.getWaitListLimit());
        summary.setWaitListProcessCode(mapper.getWaitListProcessCode());
        summary.setWaitListStatusCode(mapper.getWaitListStatusCode());
        return summary;
    }
}
